@@ -20,11 +20,14 @@
 #include <pybind11/stl.h>
 
 #include "grg_helpers.h"
+#include "grgl/common.h"
 #include "grgl/grg.h"
 #include "grgl/grgnode.h"
 #include "grgl/mutation.h"
+#include "grgl/node_data.h"
 #include "grgl/serialize.h"
 #include "grgl/ts2grg.h"
+#include "grgl/version.h"
 #include "grgl/visitor.h"
 
 namespace py = pybind11;
@@ -90,29 +93,7 @@ getTopoOrder(const grgl::GRGPtr& grg, grgl::TraversalDirection direction, const
 
 size_t hashMutation(const grgl::Mutation* self) { return std::hash<grgl::Mutation>()(*self); }
 
-std::vector<std::pair<grgl::NodeID, grgl::MutationId>> getNodeMutationPairs(const grgl::GRGPtr& grg) {
-    std::vector<std::pair<grgl::NodeID, grgl::MutationId>> result;
-    for (const auto& nodeAndMutId : grg->getNodeMutationPairs()) {
-        result.emplace_back(nodeAndMutId.first, nodeAndMutId.second);
-    }
-    return std::move(result);
-}
-
 PYBIND11_MODULE(_grgl, m) {
-    py::class_<grgl::GRG::NodeListIterator>(m, "GRG_NodeListIterator")
-        .def(
-            "__iter__",
-            [](const grgl::GRG::NodeListIterator& nli) { return py::make_iterator(nli.begin(), nli.end()); },
-            py::keep_alive<0, 1>() /* Essential: keep object alive while iterator exists */);
-
-    py::class_<grgl::NodeData>(m, "NodeData")
-        .def_readonly("num_individual_coals", &grgl::NodeData::numIndividualCoals, R"^(
-            The number of individuals that coalesce at this node.
-            )^")
-        .def_readwrite("population_id", &grgl::NodeData::populationId, R"^(
-            The population ID (integer) for the population associated with this node.
-            )^");
-
     py::class_<grgl::Mutation>(m, "Mutation")
         .def(py::init<double, std::string, const std::string&, double>(),
              py::arg("position"),
@@ -223,14 +204,6 @@ PYBIND11_MODULE(_grgl, m) {
                 :return: The parents of the given node as a list of NodeIDs.
                 :rtype: List[int]
             )^")
-        .def("get_node_data", &grgl::GRG::getNodeData, R"^(
-                Get the NodeData object associated with the NodeID.
-
-                :param node_id: The NodeID to get data for.
-                :type node_id: int
-                :return: The NodeData object.
-                :rtype: pygrgl.NodeData
-            )^")
         .def("get_sample_nodes", &grgl::GRG::getSampleNodes, R"^(
                 Get the NodeIDs for the sample nodes.
 
@@ -243,7 +216,7 @@ PYBIND11_MODULE(_grgl, m) {
                 :return: The list of NodeIDs that are root nodes.
                 :rtype: List[int]
             )^")
-        .def("get_node_mutation_pairs", &getNodeMutationPairs, R"^(
+        .def("get_node_mutation_pairs", &grgl::GRG::getNodeMutationPairs, R"^(
                 Get a list of pairs (NodeID, MutationID). Each Mutation typically
                 is associated to a single Node, but rarely it can have more than one
                 Node, in which case it will show up in more than one pair.
@@ -324,6 +297,43 @@ PYBIND11_MODULE(_grgl, m) {
 
                 :return: The mutation count.
                 :rtype: int
+            )^")
+        .def("get_population_id", &grgl::GRG::getPopulationId, py::arg("node_id"), R"^(
+                Get the population ID for the given node. Can be used to index into the list returned by
+                get_populations().
+
+                :param node_id: The node to retrieve.
+                :type node_id: int
+                :return: The population ID.
+                :rtype: int
+            )^")
+        .def("set_population_id", &grgl::GRG::setPopulationId, py::arg("node_id"), py::arg("population_id"), R"^(
+                Set the population ID for the given node.
+
+                :param node_id: The node to access.
+                :type node_id: int
+                :param population_id: The population ID to associate with the node.
+                :type population_id: int
+            )^")
+        .def("get_num_individual_coals", &grgl::GRG::getNumIndividualCoals, py::arg("node_id"), R"^(
+                Get the number of individuals that coalesced at the given node (not below or above).
+
+                :param node_id: The node to retrieve.
+                :type node_id: int
+                :return: The number of individuals that coalesced, or pygrgl.COAL_COUNT_NOT_SET.
+                :rtype: int
+            )^")
+        .def("set_num_individual_coals",
+             &grgl::GRG::setNumIndividualCoals,
+             py::arg("node_id"),
+             py::arg("num_coals"),
+             R"^(
+                Set the number of individuals that coalesced at the given node (not below or above).
+
+                :param node_id: The node to access.
+                :type node_id: int
+                :param num_coals: The number of individuals that coalesced, or pygrgl.COAL_COUNT_NOT_SET.
+                :type num_coals: int
             )^");
     grgClass.doc() = "A Genotype Representation Graph (GRG) representing a particular dataset. "
                      "This is the immutable portion of the API, so every graph has these operations. "
@@ -341,11 +351,14 @@ PYBIND11_MODULE(_grgl, m) {
 
     py::class_<grgl::MutableGRG, std::shared_ptr<grgl::MutableGRG>>(m, "MutableGRG", grgClass)
         .def(py::init<size_t, size_t>(), R"^()^")
-        .def("make_node", &grgl::MutableGRG::makeNode, py::arg("count") = 1, R"^(
+        .def("make_node", &grgl::MutableGRG::makeNode, py::arg("count") = 1, py::arg("force_ordered") = false, R"^(
             Create one or more new nodes in the graph.
 
             :param count: How many nodes to create (optional, default to 1).
             :type count: int
+            :param force_ordered: Set to True if you are sure adding this node will maintain the topological
+                order of the NodeIDs within the graph.
+            :type count: bool
         )^")
         .def("connect", &grgl::MutableGRG::connect, py::arg("source"), py::arg("target"), R"^(
             Add a down edge from source to target, and an up edge from target to source.
@@ -381,7 +394,6 @@ PYBIND11_MODULE(_grgl, m) {
           &grgl::loadImmutableGRG,
           py::arg("filename"),
           py::arg("load_up_edges") = true,
-          py::arg("load_down_edges") = true,
           R"^(
         Load a GRG file from disk. Immutable GRGs are much faster to traverse than mutable
         GRGs and take up less RAM, so this is the preferred method if you are using a GRG
@@ -391,8 +403,6 @@ PYBIND11_MODULE(_grgl, m) {
         :type filename: str
         :param load_up_edges: If False, do not load the graph "up" edges (saves RAM).
         :type load_up_edges: bool
-        :param load_down_edges: If False, do not load the graph "down" edges (saves RAM).
-        :type load_down_edges: bool
         :return: The GRG.
         :rtype: pygrgl.GRG
     )^");
@@ -547,7 +557,12 @@ PYBIND11_MODULE(_grgl, m) {
     )^");
 
     m.attr("INVALID_NODE") = INVALID_NODE_ID;
-    m.attr("COAL_COUNT_NOT_SET") = grgl::NodeData::COAL_COUNT_NOT_SET;
+    m.attr("COAL_COUNT_NOT_SET") = grgl::COAL_COUNT_NOT_SET;
 
-    m.attr("__version__") = "dev";
+    std::stringstream versionString;
+    versionString << GRGL_MAJOR_VERSION << "." << GRGL_MINOR_VERSION;
+    m.attr("__version__") = versionString.str();
+    std::stringstream fileVersionString;
+    fileVersionString << grgl::GRG_FILE_MAJOR_VERSION << "." << grgl::GRG_FILE_MINOR_VERSION;
+    m.attr("GRG_FILE_VERSION") = fileVersionString.str();
 }
